#include "mess822.h"
#include "str.h"
#include "stralloc.h"
#include "logmsg.h"

#define WHO "mess822_base64"

static uint8 *b64alpha =
  "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

static const unsigned char *b64alphaurl =
  "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_";

static unsigned int base64_dec(int c)
{
  if ('A' <= c && c <= 'Z') return c - 'A';
  if ('a' <= c && c <= 'z') return 26 + c - 'a';
  if ('0' <= c && c <= '9') return 52 + c - '0';
  if (c == '+') return 62; 
  if (c == '/') return 63; 
  return 64; 
}

/* returns number of bytes written; -1 problem 
   stralloc *out => null terminated 
   Some code borrowed from Bert Bos @ w3c.org
   RFC 4648 is considered.
*/

int mess822_b64decode(stralloc *out,const char *in,int len,int flag)
{
  int i = 0;       // FIXME: can't be done in for loop
  int r = 0;
  uint8 b;
  unsigned int c, v;

  if (!stralloc_copys(out,"")) return -1;
  if (len == 0) return 0;
  
  if (!stralloc_ready(out,len)) return -1;

  for (int remaining  = 0; i < len; i++) {
    if (in[i] == '\0') break;
    if (in[i] == '=') continue;
    if (in[i] == '\n') continue;
    if (flag == 4648) {
      if (in[i] == '_') { if (!stralloc_append(out,"/")) return -1; r++; continue; }
      if (in[i] == '-') { if (!stralloc_append(out,"+")) return -1; r++; continue; }
    }
    if ((v = base64_dec(in[i])) < 64) {
      switch (remaining) {
        case 0: c = v; remaining = 6; break;
        case 6: b = (c << 2) | (v >> 4);
                if (!stralloc_append(out,&b)) return -1;
                c = (v & 0x0f); remaining = 4; r++; break;
        case 4: b = (c << 4) | (v >> 2);
                if (!stralloc_append(out,&b)) return -1;
                c = (v & 0x03); remaining = 2; r++; break;
        case 2: b = (c << 6) | v;
                if (!stralloc_append(out,&b)) return -1;
                c = 0; remaining = 0; r++; break;
      } 
    } else if (flag == 1) { 
      char ch[] = { '\0', '\0' };
      ch[0] = in[i];
      logmsg(WHO,0,INFO,B("Erroneous character on input: >",ch,"<")); /* be quiet */
      return 0;
    }
  }
  
  return r; 
}

int mess822_b64encode(stralloc *out,stralloc *in)
{
  uint8 a, b ,c;
  uint8 b64[4];
  int i, r = 0;              // number of base64 chars written

  if (in->len == 0) {
    if (!stralloc_cats(out,"")) return -1;
    return 1;
  }

  // Compute size, mod(3) remaining bytes, #loops; allocate memory

  int len = in->len % 3; 
  int blen = in->len - len;
  int lines = (4 * in->len / 3) / 76 + 1;
  if (!stralloc_ready(out,4*in->len/3 + 2 + lines)) return -1; 

  for (i = 0; i < blen; i += 3) {
    a = in->s[i];
    b = in->s[i + 1];
    c = in->s[i + 2];

    b64[0] = b64alpha[a >> 2];
    b64[1] = b64alpha[((a & 0x03) << 4) | (b >> 4)];
    b64[2] = b64alpha[((b & 0x0f) << 2) | (c >> 6)];
    b64[3] = b64alpha[c & 0x3f];
    if (!stralloc_catb(out,b64,4)) return -1;
    r += 4;
    if (r % 76 == 0) if (!stralloc_append(out,"\n")) return -1;
  } 

  if (len == 2) {            // last two input bytes - padding
    a = in->s[i];
    b = in->s[i + 1];

    b64[0] = b64alpha[a >> 2];
    b64[1] = b64alpha[((a & 0x03) << 4) | (b >> 4)];
    b64[2] = b64alpha[b & 0x0f << 2];
    b64[3] = '=';
    if (!stralloc_catb(out,b64,4)) return -1;
    r += 4;
  } else if (len == 1) {     // last input byte - padding
    a = in->s[i];

    b64[0] = b64alpha[a >> 2];
    b64[1] = b64alpha[(a & 0x03) << 4];
    b64[2] = '=';
    b64[3] = '=';
    if (!stralloc_catb(out,b64,4)) return -1;
    r += 4;
  }

  return r;
}

int mess822_b64urlencode(stralloc *out,stralloc *in)
{
  uint8 a, b ,c;
  uint8 b64[4];
  int i, r = 0;              // number of base64 chars written

  if (in->len == 0) {
    if (!stralloc_cats(out,"")) return -1;
    return 1;
  }

  // Compute size, mod(3) remaining bytes, #loops; allocate memory

  int len = in->len % 3; 
  int blen = in->len - len;
  int lines = (4 * in->len / 3) / 76 + 1;
  if (!stralloc_ready(out,4*in->len/3 + 2 + lines)) return -1; 

  for (i = 0; i < blen; i += 3) {
    a = in->s[i];
    b = in->s[i + 1];
    c = in->s[i + 2];

    b64[0] = b64alphaurl[a >> 2];
    b64[1] = b64alphaurl[((a & 0x03) << 4) | (b >> 4)];
    b64[2] = b64alphaurl[((b & 0x0f) << 2) | (c >> 6)];
    b64[3] = b64alphaurl[c & 0x3f];
    if (!stralloc_catb(out,b64,4)) return -1;
    r += 4;
    if (r % 76  == 0) if (!stralloc_append(out,"\n")) return -1;
  } 

  if (len == 2) {            // last two input bytes - padding
    a = in->s[i];
    b = in->s[i + 1];

    b64[0] = b64alphaurl[a >> 2];
    b64[1] = b64alphaurl[((a & 0x03) << 4) | (b >> 4)];
    b64[2] = b64alphaurl[b & 0x0f << 2];
    b64[3] = '=';
    if (!stralloc_catb(out,b64,4)) return -1;
    r += 4;
  } else if (len == 1) {     // last input byte - padding
    a = in->s[i];

    b64[0] = b64alphaurl[a >> 2];
    b64[1] = b64alphaurl[(a & 0x03) << 4];
    b64[2] = '=';
    b64[3] = '=';
    if (!stralloc_catb(out,b64,4)) return -1;
    r += 4;
  }

  return r;
}
