#include "mess822.h"
#include "str.h"

static const char qptab[] =
{
//   0   1   2   3   4   5   6    7   8   9   10   11   12  13  14  15
     0,  0,  0,  0,  0,  0,  0,   0,  0,'\t','\n',  0,   0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,   0,  0,  0 ,  0 ,  0,   0,  0,  0,  0,
  ' ' ,'!','"','#','$','%','&','\'','(',')' ,'*' ,'+', ',','-','.','/',
  '0' ,'1','2','3','4','5','6','7' ,'8','9' ,':' ,';', '<',  0,'>','?',
  '@' ,'A','B','C','D','E','F','G' ,'H','I' ,'J' ,'K', 'L','M','N','O',
  'P' ,'Q','R','S','T','U','V','W' ,'X','Y' ,'Z' ,'[','\\',']','^','_',
  '`' ,'a','b','c','d','e','f','g' ,'h','i' ,'j' ,'k', 'l','m','n','o',
  'p' ,'q','r','s','t','u','v','w' ,'x','y' ,'z' ,'{', '|','}','~',  0,
};

static const char hextab[] = "0123456789abcdef";
static const char qpok[] = "0123456789abcdef ";  // FIXME: Invalid last char required
static const char QPok[] = "0123456789ABCDEF ";  // FIXME: Invalid last char required

int hex2dec(char hex[2])
{
  unsigned char c, dec = 0;
  
  for (int i = 0; i < 2; i++) {
    c = hex[i];
    dec <<= 4;
    if (c >= '0' && c <= '9') dec += c - '0';
    else if (c >= 'A' && c <= 'F') dec += c - 55; // 'A' - 10
    else if (c >= 'a' && c <= 'f') dec += c - 87; // 'a' - 10
    else return -1; 
  }   
  return dec;
}

static int decode_qp(const char *s)
{
  if (*s == '=') {
    if (*s == '\n') s + 1;
    return hex2dec(s + 1);
  } else 
    return -1;
}

/* If the character is <= ' ', > '~', or '=', we want to
 * print it out in =XY format, unless it's a \n, where
 * we dump a RFC822 \r\n, or unless it's a tab or space,
 * where we display it as is, except if it's at end of
 * line we need to dump a soft linebreak to keep from
 * having a bare whitespace at the end of the line
 * (RFC1541 5.1, rule 3).  If the line ever gets over
 * 70 characters long, we need to drop a soft linebreak
 * and start a new line (RFC1541 5.1, rule 5).  If none
 * of the above is true, dump the ascii value of the
 * character and continue on our way.
 *
 * Return number of byte encoded
*/

int mess822_qpencode(stralloc *out,stralloc *in)
{
  int r = 0;       // number of written chars
  unsigned char ch;
  char xch[2];

  if (in->len == 0) {
    if (!stralloc_copys(out,"")) return -1;
    return 0;
  }
  if (!stralloc_copys(out,"")) return -1;

  for (int i = 0; i < in->len - 1; i++) {
    ch = in->s[i];
    if ((ch & 0x80) || (ch == '.') || (qptab[ch] == 0)) {
      if (!stralloc_append(out,"=")) return -1;
      r++;
      xch[0] = hextab[(ch >> 4) & 0x0f];
      xch[1] = hextab[(ch & 0x0f)];
      if (!stralloc_catb(out,xch,2)) return -1; 
      r += 2;
    } else {
      if (!stralloc_append(out,in->s + i)) return -1; 
      r++;
    }
    if (r % 76 == 0) if (!stralloc_catb(out,"=\n",2)) return -1;
    r += 2;
  }

  return r;
}

/* Decode a quoted-printable string:
   1. Convert 2 byte chars introduced by '='.
   2. Respect \r\n and \r.
   3. Translate '_' as WSP if RFC 2047.
   4. Treat standard chars unchanged.
   The incoming byte sequence does not need to be \0 terminated.
   The outgoing stralloc is \0 terminated.
 
   Return number of bytes decoded, -1 if out of memory
*/

int mess822_qpdecode(stralloc *out,const char *in,int len,int flag)
{
  int r = 0;
  char ch;

  if (len == 0) {
    if (!stralloc_copys(out,"")) return -1;
    return 1;
  }
  if (!stralloc_readyplus(out,len)) return -1;

  while (len) {
    if (*in == '\0') return r;         // done
    if (flag == 2047 && *in == '_') {  // RFC 2047 § 4.2(2)
      if (!stralloc_append(out," ")) return -1; 
      r++, in++; len--; continue; 
    }
    if (*in == '=')  {                 // QP double byte char
      if (*(in + 1) == '\n') { in += 2; len -= 2; continue; }
      if ((str_chr(QPok,*(in + 1)) < 16 && str_chr(QPok,*(in + 2)) < 16) ||
          (str_chr(qpok,*(in + 1)) < 16 && str_chr(qpok,*(in + 2)) < 16)) {
        ch = hex2dec(in + 1);
        if (!stralloc_append(out,&ch)) return -1;
        r++; in += 3; len -= 3;
      } else { r += 2; in += 2; len -= 2; continue; }      // ignore any none-valid char
    }
    else {                             // standard char
     if (!stralloc_append(out,in)) return -1; 
     r++; in++; len--; 
    } 
  }
  if (!stralloc_0(out)) return -1;
 
  return r;
}
